from langchain_openai import ChatOpenAI, OpenAIEmbeddings
from datasets import Dataset
from ragas import evaluate
from ragas.metrics import (
    answer_relevancy,
    answer_similarity,
    answer_correctness
)
import pandas as pd
from ragas.testset.generator import TestsetGenerator
import yaml
import pandas as pd
from ragas.testset.generator import TestsetGenerator
from ragas.testset.evolutions import simple, multi_context
import argparse

class SyntheticDataGenerator:
    """
    A class for generating synthetic test datasets using language models.

    Args:
        documents (list): A list of documents to generate test data from.
        config (dict): A dictionary containing configuration parameters.
    
    Methods:
        generate_testset(num_docs=8): Generates a synthetic test dataset and saves it as a CSV file.
    """

    def __init__(self, documents, config):
        
        self.documents = documents
        self.generator_llm = ChatOpenAI(model="gpt-3.5-turbo-0125",temperature=0.3)
        self.critic_llm = ChatOpenAI(model="gpt-3.5-turbo-0125",temperature=0.3)
        self.embeddings = OpenAIEmbeddings()
        self.generator = TestsetGenerator.from_langchain(
            self.generator_llm,
            self.critic_llm,
            self.embeddings,
        )
        self.save_dir = config["data"]["save_dir"]
        self.distributions = {
            simple: 0.8,
            multi_context: 0.2,
        }

    def generate_testset(self, num_docs=5):
        """
        Generates a synthetic test dataset and saves it as a CSV file.

        Args:
            num_docs (int, optional): Number of documents to generate (default is 5).
        
        Returns:
            Synthetic dataset amd saves a copy to the current working directory or any custom path. 
        """
        testset = self.generator.generate_with_langchain_docs(self.documents, num_docs, self.distributions, raise_exceptions=False)
        save_path=self.save_dir + "/synthetic_data/syntheticdataset.csv"
        testset.to_dataset().remove_columns(['contexts', 'evolution_type','episode_done','metadata']).to_csv(save_path)
        return save_path



class Generation_Benchmarking:
    """
   Benchmarks and evaluates response generators on a test dataset.

   Attributes:
       testset_df (pandas.DataFrame): Test dataset with questions, ground truth, and generator outputs.
       config (dict): yaml file containing relevant instructions  
    """
    def __init__(self, testset_df, config):
        self.testset_df = testset_df
        metrics = [answer_relevancy, answer_similarity, answer_correctness]
        config = config['generator']['generation_benchmark_metrics']
        self.filtered_metrics = []
        for metric in metrics:
            if config[metric.name]:
                self.filtered_metrics.append(metric)

    def run_benchmarks(self):
        """
       Evaluate response generators and return the best performer with scores.

       Returns:
           dict: Best generator column and its evaluation scores.

       Raises:
           ValueError: If required columns are missing.
        """
            
        required_columns = ['question', 'ground_truth', 'contexts']

        response_columns = [col for col in self.testset_df.columns if col not in required_columns]

        if not set(required_columns).issubset(set(self.testset_df.columns)):
            raise ValueError("The required columns 'question', 'ground_truth', and 'contexts' are missing.")
        dataset = Dataset.from_pandas(self.testset_df)
        generator_benchmarks = {}

        for col in response_columns:
            dataset = dataset.rename_column(col, 'answer')
            generator_benchmarks[col] = evaluate(dataset, metrics=self.filtered_metrics, raise_exceptions=False)
            dataset = dataset.remove_columns('answer')
        

        average_scores = {}
        for combination in generator_benchmarks.keys():
            avg_score = sum(generator_benchmarks[combination].values()) / len(generator_benchmarks[combination])
            average_scores[combination] = avg_score

        best_combination = max(average_scores, key=average_scores.get)

        return {best_combination: generator_benchmarks[best_combination]}

if __name__=="__main__":

    parser = argparse.ArgumentParser(description='Generation Benchmarking')

    parser.add_argument('--testset_file', type=str, required=True,
                        help='Path to the testset file')

    args = parser.parse_args() 
    
    save_dir = pd.read_csv(args.testset_file)

    with open("config.yaml", 'r') as file:
        conf = yaml.load(file, Loader=yaml.FullLoader)

    if args.testset_file:
        conf['data']['benchmark_data'] =  args.config_file
        
    gen_bench = Generation_Benchmarking(testset_df=save_dir, config=conf).run_benchmarks()
    
    print(gen_bench)



import os
import random
import chardet
import csv
import yaml
import nltk
from PyPDF2 import PdfReader
from nltk.corpus import stopwords
from nltk.tokenize import word_tokenize
from nltk.stem import WordNetLemmatizer
from langchain.text_splitter import CharacterTextSplitter,RecursiveCharacterTextSplitter
import argparse

# Download necessary NLTK resources
nltk.download('wordnet')
nltk.download('stopwords')
nltk.download('punkt')
class DataProcessor:
    def __init__(self, config):
        # Initialize class variables
        self.config = config
        self.stopwords = set(stopwords.words("english"))
        self.lemmatizer = WordNetLemmatizer()
        self.text_splitter = RecursiveCharacterTextSplitter(
            # separator=" ",
            chunk_size=self.config["retriever"]["chunk_size"],
            chunk_overlap=self.config["retriever"]["text_overlap"]
        )

    def process_data(self):
        # Process data from file paths in the configuration file
        corpus_file_paths = self.config["data"]["corpus"]

        chunks = []
        for file_path in corpus_file_paths:
            # Determine file type and read text
            file_type = os.path.splitext(file_path)[1][1:]
            if file_type.lower() not in ['pdf', 'csv', 'txt']:
                print(f"Warning: Unsupported file type: {file_path}")
                continue
            if file_type.lower() == 'pdf':
                with open(file_path, "rb") as f:
                    pdf = PdfReader(f)
                    text = ""
                    for page in pdf.pages:
                        text += page.extract_text()
            elif file_type.lower() == 'csv':
                with open(file_path, "r") as f:
                    reader = csv.reader(f)
                    text = "\n".join([",".join(row) for row in reader])
            else:
                with open(file_path, "rb") as f:
                    rawdata = f.read()
                result = chardet.detect(rawdata)
                encoding = result['encoding']
                with open(file_path, "r", encoding=encoding) as f:
                    text = f.read()
            # Preprocess text and split into chunks
            processed_text = self.process_text(text)
            chunks.extend(self.split_into_chunks(processed_text))

        return chunks

    def process_text(self, text):
        # Preprocess text by tokenizing, removing stopwords, and lemmatizing
        tokens = word_tokenize(text)
        tokens = [token.lower() for token in tokens if token.isalnum()]
        tokens = [token for token in tokens if token not in self.stopwords]
        tokens = [self.lemmatizer.lemmatize(token) for token in tokens]
        return " ".join(tokens)

    def split_into_chunks(self, data):
        # Split text into chunks using the CharacterTextSplitter
        chunks = self.text_splitter.split_text(data)
        return chunks


if __name__ == "__main__":
    # Parse command line arguments
    parser = argparse.ArgumentParser(description="Data Processing")
    parser.add_argument("--config", type=str, default="./config.yaml",help="Path to the configuration file")
    parser.add_argument("--user_files", nargs='+', type=str, default=None, help="Path to the user-specified file to be processed")
    parser.add_argument("--chunk_size", type=int, default=400, help="Chunk size for splitting text")
    parser.add_argument("--text_overlap", type=int, default=50, help="Text overlap for splitting text")
    args = parser.parse_args()

    # Load configuration file
    with open(args.config, "r") as f:
        config = yaml.safe_load(f)

    # Update configuration if user provides file path
    if args.user_files:
        config["data"]["corpus"] = args.user_files

    # Update configuration if user provides chunk size
    if args.chunk_size:
        config["retriever"]["chunk_size"] = args.chunk_size

    # Update configuration if user provides text overlap
    if args.text_overlap:
        config["retriever"]["text_overlap"] = args.text_overlap

    # Create instance of DataProcessor class
    processor = DataProcessor(config)

    # Call process_data method to get chunks
    chunks = processor.process_data()




from langchain.docstore.document import Document
from langchain_community.embeddings.sentence_transformer import SentenceTransformerEmbeddings
from langchain_text_splitters import CharacterTextSplitter
from sentence_transformers import SentenceTransformer
from transformers import AutoModel, AutoTokenizer
from langchain_community.embeddings import HuggingFaceEmbeddings
import torch
import torch.nn.functional as F
import os
import itertools
from langchain_community.vectorstores import Chroma
from langchain_community.vectorstores import FAISS
from langchain_community.embeddings import HuggingFaceInstructEmbeddings
# from sklearn.metrics.pairwise import cosine_similarity
# import numpy as np
from langchain_openai import OpenAIEmbeddings
import yaml
# from dotenv import load_dotenv
# load_dotenv()

# os.environ["OPENAI_API_KEY"] = os.getenv("OPENAI_API_KEY")
from itertools import product
from tqdm import tqdm
import argparse

class EmbeddingGenerator:
    def __init__(self, config):
        self.config = config
        self.openai_api_key = os.environ.get("OPENAI_API_KEY")

    def huggingface_instruct_embeddings(self, chunks, vectorstore):
        docs = []
        for chunk in chunks:
            docs.append(
                Document(
                    page_content=chunk
                )
            )
        if not docs:
            return "Chunks list is empty"
        embedding_function = HuggingFaceInstructEmbeddings()
        if vectorstore == "Faiss":
            persist_directory = os.path.join(self.config["retriever"]["vector_store"]["persist_directory"][0], "huggingface_instruct_embeddings_faiss")
            db = FAISS.from_documents(docs, embedding_function)
            db.save_local(persist_directory)
        else:
            persist_directory = os.path.join(self.config["retriever"]["vector_store"]["persist_directory"][0], "huggingface_instruct_embeddings_chroma")
            db = Chroma(persist_directory=persist_directory, embedding_function=embedding_function)
            db.add_documents(docs)
        return db

    def all_minilm_embeddings(self, chunks, vectorstore):
        docs = []
        for chunk in chunks:
            docs.append(
                Document(
                    page_content=chunk
                )
            )
        if not docs:
            return "Chunks list is empty"
        embedding_function = SentenceTransformerEmbeddings(model_name="all-MiniLM-L6-v2")
        if vectorstore== "Faiss":
            persist_directory = os.path.join(self.config["retriever"]["vector_store"]["persist_directory"][0], "all_minilm_embeddings_faiss")
            db = FAISS.from_documents(docs, embedding_function)
            db.save_local(persist_directory)
        else:
            persist_directory = os.path.join(self.config["retriever"]["vector_store"]["persist_directory"][0], "all_minilm_embeddings_chroma")
            db = Chroma(persist_directory=persist_directory, embedding_function=embedding_function)
            db.add_documents(docs)
        return db
    

    def bgem3_embeddings(self, chunks, vectorstore):
        docs = []
        for chunk in chunks:
            docs.append(
                Document(
                    page_content=chunk
                )
            )
        if not docs:
            return "Chunks list is empty"
        embedding_function = HuggingFaceEmbeddings(model_name="BAAI/bge-large-en-v1.5")
        if vectorstore == "Faiss":
            persist_directory = os.path.join(self.config["retriever"]["vector_store"]["persist_directory"][0], "bgem3_embeddings_faiss")
            db = FAISS.from_documents(docs, embedding_function)
            db.save_local(persist_directory)
        else:
            persist_directory = os.path.join(self.config["retriever"]["vector_store"]["persist_directory"][0], "bgem3_embeddings_chroma")
            db = Chroma(persist_directory=persist_directory, embedding_function=embedding_function)
            db.add_documents(docs)
        return db


    def openai_embeddings(self, chunks, vectorstore):
        docs = []
        for chunk in chunks:
            docs.append(
                Document(
                    page_content=chunk
                )
            )
        if not docs:
            return "Chunks list is empty"
        if self.openai_api_key:
            embedding_function = OpenAIEmbeddings(openai_api_key=self.openai_api_key)
            if vectorstore == "Faiss":
                persist_directory = os.path.join(self.config["retriever"]["vector_store"]["persist_directory"][0], "openai_embeddings_faiss")
                db = FAISS.from_documents(docs, embedding_function)
                db.save_local(persist_directory)
            else:
                persist_directory = os.path.join(self.config["retriever"]["vector_store"]["persist_directory"][0], "openai_embeddings_chroma")
                db = Chroma(persist_directory=persist_directory, embedding_function=embedding_function)
                db.add_documents(docs)
            return db
        else:
            print("OpenAI API key is not set in the environment variable.")
            return None
    
    def generate_databases(self, chunks):
        embedding_methods = self.config["retriever"]["vector_store"]["embedding"]
        vectorstore_option = self.config["retriever"]["vector_store"]["database"]
        databases = []
        permutations = list(itertools.product(embedding_methods, vectorstore_option))
        for embedding_method,vectorstore_option in permutations:
            # print("creating db with embeddings:{} and vector store:{}".format(embedding_method,vectorstore_option))
            try:
                if embedding_method == "huggingface_instruct_embeddings":
                    db = self.huggingface_instruct_embeddings(chunks, vectorstore_option)
                elif embedding_method == "all_minilm_embeddings":
                    db = self.all_minilm_embeddings(chunks, vectorstore_option)
                elif embedding_method == "bgem3_embeddings":
                    db = self.bgem3_embeddings(chunks, vectorstore_option)
                elif embedding_method == "openai_embeddings":
                    db = self.openai_embeddings(chunks, vectorstore_option)
                else:
                    raise ValueError("Invalid embedding method specified.")

                di = {
                    "embeddings": embedding_method,
                    "vectorstore": vectorstore_option,
                    "db": db
                }
                databases.append(di)
            except Exception as e:
                print(f"Error generating database for {embedding_method} and {vectorstore_option}: {e}")

        return databases

if __name__ == "__main__":
    config_file = "./config.yaml"
    with open(config_file, 'r') as f:
        config = yaml.safe_load(f)

    parser = argparse.ArgumentParser()
    parser.add_argument("--embedding", nargs='+', help="List of embedding options available are: huggingface_instruct_embeddings, all_minilm_embeddings, bgem3_embeddings, openai_embeddings")
    parser.add_argument("--vectorstore", nargs='+', type=str, help="Vector store option (chroma or faiss)")
    parser.add_argument("--chunks", nargs='+', help="List of text chunks")
    args = parser.parse_args()

    if args.embedding:
        config["retriever"]["vector_store"]["embedding"] = args.embedding

    if args.vectorstore:
        config["retriever"]["vector_store"]["database"] = args.vectorstore

    if args.chunks:
        config["retriever"]["vector_store"]["chunks"] = args.chunks

    obj = EmbeddingGenerator(config)
    dbs = obj.generate_databases(config["retriever"]["vector_store"]["chunks"])

    for db in dbs:
        print(db)




import yaml 
import pandas as pd
import os,ast
import pickle
from datasets import Dataset
from ragas import evaluate
from statistics import mean 
import argparse
from ragas.testset.evolutions import simple, reasoning, multi_context
from ragas.metrics import (
    context_precision,
    context_recall,
)

class RetrievalBenchmarking:
    def __init__(self,config,datasets_dir_path):
        self.config = config
        d={}
        
        for df_name in os.listdir(datasets_dir_path):
          if ".csv" in df_name:
            d[df_name]=pd.read_csv(os.path.join(datasets_dir_path,df_name))

        self.dict_data = d
    
    def validate_dataframe(self):
        """
        Validates the format of the input DataFrame `df_retrieval`.
        
        Performs benchmarking using RAGAS
        """
        
        max_combo = None
        max_benchmark_avg = 0
        for key, df_retrieval in self.dict_data.items():
            default_metrics = [context_precision,context_recall]        
            df_retrieval['contexts'] = df_retrieval['contexts'].apply(ast.literal_eval)
            config_metrics = self.config['retriever']['retriever_benchmark_metrics']
            filtered_metrics = []

            for metric in default_metrics:
                if config_metrics[metric.name]:
                    filtered_metrics.append(metric)

            # Check if the input is a DataFrame
            if not isinstance(df_retrieval, pd.DataFrame):
                print("Input is not a pandas DataFrame.")
                return False
            
            # Check if the required columns exist
            required_columns = ["question", "ground_truth","contexts"]
            if not set(required_columns).issubset(df_retrieval.columns):
                print("DataFrame is missing one or more required columns.")
                return False
            testsetdf = Dataset.from_pandas(df_retrieval)
            result = evaluate(
                    dataset=testsetdf, 
                    metrics = filtered_metrics,
                    )
            scores_list = []
            for metric, score in result.items(): 
                scores_list.append(score)
            avg = mean(scores_list)
            print("average_metrics", avg)
            if avg > max_benchmark_avg:
                max_benchmark_avg=avg
                max_combo=key

        for key,df in self.dict_data.items():
            if key == max_combo:
                return df,max_combo

             
            
if __name__=="__main__":

    parser = argparse.ArgumentParser(description='Retrieval Benchmarking')
    parser.add_argument('--retrieved_data_dir',type=str, required=True,help="folder should contain benchmarking dataset with columns as question ground truth and contexts ")                  
    parser.add_argument("--config", type=str, default="./config.yaml")

    args = parser.parse_args() 
    datasets_dir = args.retrieved_data_dir
    
    with open(args.config, 'r') as file:
        config = yaml.load(file, Loader=yaml.FullLoader)

    df,max_combo = RetrievalBenchmarking(datasets_dir_path=datasets_dir, config=config).validate_dataframe()
    print("max cobo found at{}".format(max_combo))








from langchain.docstore.document import Document
import argparse
import yaml
from ragpy.src.embeddings_creation.embedding_generator import EmbeddingGenerator
import os
from ragpy.src.dataprocessing.data_loader import DataProcessor
from ragpy.src.generator.generation_benchmarking import SyntheticDataGenerator
import pandas as pd
from sentence_transformers import CrossEncoder



class Reranking:
    def __init__(self,config):
        self.config=config
        self.cross_encoder=CrossEncoder('cross-encoder/ms-marco-MiniLM-L-6-v2')
        

    def ret(self,data,top_k, config, dict_db =None, num_questions=5):
        # df_dict={}
        fpath=""
        if self.config["data"]["benchmark_data"]:
            fpath=self.config["data"]["benchmark_data"]
        else:
            documents = [Document(content) for content in data]
            s=SyntheticDataGenerator(documents,self.config)
            fpath=s.generate_testset(num_docs=num_questions)
        
        df=pd.read_csv(fpath)
        col_list=['question','ground_truth','contexts']
        try:
            if set(col_list).issubset(set(df.columns)):
                save_dir = self.config["data"]["save_dir"] + "/retrieved_data/"
                if not os.path.exists(save_dir):
                    os.makedirs(save_dir)

                for d in dict_db:    
                    df['contexts']=''
                    # df['Generated Query']=''
                    df_name = f'{d["embeddings"]}_{d["vectorstore"]}'
                    for k in range(0,df.shape[0]):
                        query=df.iloc[k,0]
                        db=d["db"]
                        docs=db.similarity_search_with_relevance_scores(query)
                        l1=[]
                        for doc in docs:
                            l1.append(doc[0].page_content)
                        reranked_documents=self.rerank_documents(query,l1,top_k)
                        l2=[]
                        for doc in reranked_documents:
                            l2.append(doc)
                        df.at[k,'contexts']=l2
                        # df_dict[df_name] = df
                
                    file_path = save_dir + f"{df_name}.csv"
                    df.to_csv(file_path, index=False, encoding='utf-8')
                    print("Dataframe saved to", file_path)
                
                return save_dir
        except:
            raise ValueError("Dataframe should contain tcolumns in the format of 'question','ground_truth','contexts'")

    def rerank_documents(self, query, documents, top_n):
        scores={}
        for doc in documents:
          scores[doc]=self.cross_encoder.predict([query,doc])
        
        r=list(dict(sorted(scores.items(),key= lambda item:item[1],reverse=True)))[:top_n]
        
        return r 
    
if __name__ == "__main__":

    config_file='./config.yaml'
    parser = argparse.ArgumentParser(description='Retrieval and Reranking')
    parser.add_argument('--top_k',help='The number of top documents to be retrieved')
    parser.add_argument('--save_dir',help='Directory to save synthetic data')
    parser.add_argument('--benchmark_data_path',help="Path to the benchmarking dataset in csv")
    parser.add_argument('--num_questions',help="Number of questions to be generated in synthetic benchmark dataset",default=5)

    
    args = parser.parse_args()

    with open(config_file, 'r') as file: 
        config = yaml.safe_load(file)

    if args.top_k:
        config["retriever"]["top_k"] = int(args.top_k)
    if args.benchmark_data_path:
        config["data"]["benchmark_data"]=args.benchmark_data_path
    if args.save_dir:
        config["data"]["save_dir"]=args.save_dir
    
    c=DataProcessor(config) 
    chunks=c.process_data()
    obj=EmbeddingGenerator(config)
    dict_db=obj.generate_databases(chunks)

    m=Reranking(config)
    q=m.ret(chunks,config["retriever"]["top_k"],config,dict_db=dict_db, num_questions=int(args.num_questions))
    print(q)



# from setuptools import setup, find_packages

# with open('requirements.txt') as f:
#     required_packages = f.read().splitlines()

# setup(
#     name='ragpy',
#     version='0.1.0',
#     author="BayesLabs",
#     author_email="contact@bayeslabs.co",
#     packages=find_packages(),
#     py_modules=["ragpy.src.dataprocessing.data_loader",
#     "ragpy.src.embeddings_creation.embedding_generator",
#     "ragpy.src.retriever.retrieval",
#     "ragpy.src.retriever.retrieval_benchmarking",
#     "ragpy.src.generator.generation_benchmarking"],
#     install_requires = required_packages
# )
from setuptools import setup, find_packages
setup( 
    name='ragpy_new',
    version='0.1.17',                                                                                                                  
    author="BayesLabs", 
    author_email="contact@bayeslabs.co",
    packages=find_packages(),
    py_modules=[
        "ragpy.src.dataprocessing.data_loader",
        "ragpy.src.embeddings_creation.embedding_generator",
        "ragpy.src.retriever.retrieval",
        "ragpy.src.retriever.retrieval_benchmarking",
        "ragpy.src.generator.generation_benchmarking"
    ],
    install_requires=[
        "langchain==0.1.9",
        "sentence_transformers==2.2.2",
        "langchain_community==0.0.34",
        "datasets>=2.19.1",
        "langchain-openai==0.1.3",
        "langchain-fireworks==0.1.2",
        "pypdf2>=3.0.1",
        "chardet>=5.2.0",
        "nltk>=3.8.1",
        "chromadb>=0.4.24",
        "faiss-cpu>=1.8.0",
        "InstructorEmbedding>=1.0.1",
        "langchain_text_splitters>=0.0.2",
        "pandas==2.2.2"     
    ]  
) 

